﻿using System;
using System.Linq;
using System.Windows.Forms;

namespace Xompare
{
    using System.Drawing;
    using System.IO;

    using Utilities;

    using Xompare.Helpers;

    public partial class XmlFilesCompareControl : UserControl
    {
        /// <summary>
        /// Tracks the last size of the control.
        /// </summary>
        private Size m_lastSize = new Size(0, 0);

        /// <summary>
        /// Defines the event triggered when the compared files are chosen differently.
        /// </summary>
        /// <param name="newCompareName">The new comparison name.</param>
        public delegate void ComparedFilesChanged(string newCompareName);
        public event ComparedFilesChanged OnComparedFilesChanged;

        public XmlFilesCompareControl()
        {
            InitializeComponent();

            // Set viewers' roles:
            xmlFileTreeOrigin.FileViewerRole = FileComparisonRole.OriginSide;
            xmlFileTreeTarget.FileViewerRole = FileComparisonRole.TargetSide;

            Action comparedFilesChangedNotification =
                () =>
                {
                    string originalName = Path.GetFileName(NameAndLabel.GetSafeName(OriginalXmlPath));
                    string targetName = Path.GetFileName(NameAndLabel.GetSafeName(TargetXmlPath));

                    OnComparedFilesChanged(originalName + " <-> " + targetName);

                    if (!string.IsNullOrEmpty(originalName) && !string.IsNullOrEmpty(targetName))
                    {
                        // Compare the files:
                        var comparer = new CollectionComparer<string, ComparableSet<KeyedXmlRecord>>(
                            xmlFileTreeOrigin.PreParseInformation.TreeRecords, xmlFileTreeTarget.PreParseInformation.TreeRecords);

                        if (!comparer.OnlyOriginal.Any() && !comparer.OnlyTarget.Any() && !comparer.Differences.Any())
                        {
                            LogUi.LogInfo("Chosen files are identical.");
                        }

                        // Set the differences' information to the viewers:
                        xmlFileTreeOrigin.XmlOnlyHere = comparer.OnlyOriginal;
                        xmlFileTreeOrigin.XmlOnlyThere = comparer.OnlyTarget;
                        xmlFileTreeOrigin.XmlDifferences = comparer.Differences;
                        xmlFileTreeTarget.XmlOnlyHere = comparer.OnlyTarget;
                        xmlFileTreeTarget.XmlOnlyThere = comparer.OnlyOriginal;
                        xmlFileTreeTarget.XmlDifferences = comparer.Differences;

                        // Refesh the XML tree representation:
                        RefreshComparison();
                    }
                };

            xmlFileTreeOrigin.OnFilePathChanged += comparedFilesChangedNotification;
            xmlFileTreeTarget.OnFilePathChanged += comparedFilesChangedNotification;
            xmlFileTreeOrigin.OnDiffNodeSelected += xmlFileTreeTarget.EnsureNodeVisibleByXmlPath;
            xmlFileTreeTarget.OnDiffNodeSelected += xmlFileTreeOrigin.EnsureNodeVisibleByXmlPath;
        }

        /// <summary>
        /// Gets or sets the path to the XML file which is considered the original.
        /// </summary>
        public NameAndLabel OriginalXmlPath
        {
            get
            {
                return xmlFileTreeOrigin.FilePath;
            }

            set
            {
                xmlFileTreeOrigin.FilePath = value;
            }
        }

        /// <summary>
        /// Gets or sets the path to the XML file which is considered the target.
        /// </summary>
        public NameAndLabel TargetXmlPath
        {
            get
            {
                return xmlFileTreeTarget.FilePath;
            }

            set
            {
                xmlFileTreeTarget.FilePath = value;
            }
        }

        /// <summary>
        /// Refresh the comparison view.
        /// </summary>
        public void RefreshComparison()
        {
            xmlFileTreeOrigin.RefreshXmlTree();
            xmlFileTreeTarget.RefreshXmlTree();
        }

        private void XmlFilesCompareControl_Resize(object sender, EventArgs e)
        {
            if (Size.Width > 0)
            {
                var widthRatio = Size.Width / (double)m_lastSize.Width;

                splitterXmlFiles.SplitPosition = (int)((splitterXmlFiles.SplitPosition + splitterXmlFiles.Width / 2) * widthRatio) - splitterXmlFiles.Width / 2;
                m_lastSize = Size;
            }
        }

        private void splitterXmlFiles_DoubleClick(object sender, EventArgs e)
        {
            splitterXmlFiles.SplitPosition = Width / 2 - splitterXmlFiles.Width / 2;
        }

        private void XmlFilesCompareControl_Load(object sender, EventArgs e)
        {
            m_lastSize = Size;
            splitterXmlFiles.SplitPosition = Width / 2 - splitterXmlFiles.Width / 2;
        }
    }
}
